import numpy as np
import pandas as pd
from scipy.optimize import minimize
from .components import Rule

from pypfopt.expected_returns import mean_historical_return
from pypfopt.risk_models import CovarianceShrinkage
from pypfopt.efficient_frontier import EfficientFrontier


def risk_budgeting(covmat, budget):
    def objective(w, cov, rb) :
        var = w.T @ cov @ w
        sgm = var ** 0.5
        mrc = 1 / sgm * (cov @ w)
        rc = w * mrc
        rr = rc / sgm # Relative risk contribution
        return np.sum(np.square(rr - rb))

    covmat[covmat < 0] = 0  # Negative corelation is adjusted to zero.

    cnt = covmat.shape[0]
    w0 = [1 / cnt] * cnt
    constraints = ({'type': 'eq', 'fun': lambda x: x.sum() - 1},
                {'type': 'ineq', 'fun': lambda x: x})
    options = {'ftol': 1e-20, 'maxiter': 10000}
    result = minimize(fun=objective,
                    x0=w0,
                    args=(covmat, budget),
                    method='SLSQP',
                    constraints=constraints,
                    options=options)
    return(result.x)


class Pipeline(Rule):
    def __init__(self, rules):
        self.rules = rules

    def calculate(self, date, universe, fund):
        assets = self.rules[0].assets
        for rule in self.rules:
            rule.set_assets(assets)
            weights = rule.calculate(date, universe, fund)
            assets = weights.keys().tolist()

        return weights


class ConstantWeight(Rule):
    def __init__(self, weights):
        self.weights = pd.Series(weights)

        # assert sum(self.weights) == 1

    def calculate(self, date, universe, fund):
        return self.weights


class EqualWeight(Rule):
    def __init__(self, assets):
        n_assets = len(assets)
        weights = 1 / n_assets
        self.weights = pd.Series([weights] * n_assets, index=assets)

    def calculate(self, date, universe, fund):
        return self.weights

 
class RiskParity(Rule):
    def __init__(self, assets=None, window=252):
        self.assets = assets
        self.window = window
    
    def set_assets(self, assets):
        self.assets = assets

    def calculate(self, date, universe, fund):
        n_assets = len(self.assets)
        returns = universe.pricing['return'].unstack()[self.assets]
        returns = returns.loc[:date].iloc[-self.window:]
        covmat = returns.cov().values
        budget = [1 / n_assets] * n_assets 
        weights = risk_budgeting(covmat, budget)
        weights = pd.Series(weights, index=self.assets)
        return weights


class TopNbyMomentum(Rule):
    def __init__(self, assets=None, top_n=1, period=126):
        self.assets = assets
        self.top_n = top_n
        self.period = period

    def set_assets(self, assets):
        self.assets = assets

    def calculate(self, date, universe, fund):
        prices = universe.pricing['price'].unstack()[self.assets]
        momentums = prices.iloc[-1] / prices.iloc[-self.period] - 1
        selected = momentums.sort_values(ascending=False).iloc[:self.top_n].index
        weight = 1 / len(selected)
        weights = pd.Series([weight] * len(selected), index=selected)
        return weights


class MinimumVariance(Rule):
    def __init__(self, assets=None, window=252):
        self.assets = assets
        self.window = window

    def set_assets(self, assets):
        self.assets = assets

    def calculate(self, date, universe, fund):
        prices = universe.pricing['price'].unstack()[self.assets].iloc[-self.window:]
        mu = mean_historical_return(prices)
        S = CovarianceShrinkage(prices).ledoit_wolf()
        ef = EfficientFrontier(mu, S)
        weights = ef.min_volatility()
        weights = pd.Series(weights, index=self.assets)
        return weights



from .backtest import (
    run_backtest
)

from .components import (
    Portfolio,
    Universe,
    Rule,
)

from .rules import (
    EqualWeight,
    RiskParity,
    ConstantWeight,
    Pipeline,
    TopNbyMomentum,
    MinimumVariance,
)

from .report import (
    calc_stats,
    report_perf,
)

from . import _version
__version__ = _version.get_versions()['version']



import pandas as pd
import pytest
import os
import snowball as sb

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(TEST_DIR, 'data', 'etfs_prices.csv')

@pytest.fixture(scope="module")
def sample_prices():
    print("\n--- Fixture: Starting sample_prices load ---")
    prices = pd.read_csv(DATA_PATH, index_col=0, parse_dates=True)
    prices.index.name = 'Date'
    prices = prices.rename_axis(None, axis=1)
    print("--- Fixture: sample_prices load complete ---")
    return prices

def test_backtest(sample_prices):
    print("--- Test: test_backtest running ---")
    bt = sb.run_backtest(prices=sample_prices, 
                         schedule='EOM',
                         rule={'069500': 0.6, '114820': 0.4},
                         cost=0.002,
                         start='2020-01-01', end='2024-12-31')
    actual_value = bt.stats['CAGR']
    expected_value = -0.03308272002801749
    assert actual_value == pytest.approx(expected_value, abs=1e-10), \
        f"Backtest result is not {expected_value}. Actual result: {actual_value}"
    print("--- Test: test_backtest completed ---")

def test_calc_stats(sample_prices):
    print("--- Test: test_calc_stats running ---")
    dummy_returns = sample_prices['069500'].pct_change().iloc[1:]
    stats = sb.calc_stats(dummy_returns)
    actual_value = stats['CAGR']
    expected_value = -0.1231127835849668
    assert actual_value == pytest.approx(expected_value, abs=1e-10), \
        f"Backtest result is not {expected_value}. Actual result: {actual_value}"
    print("--- Test: test_calc_stats completed ---")



from abc import ABCMeta, abstractmethod
from sqlite3 import Timestamp
import numpy as np
import pandas as pd
import math
from tqdm.auto import tqdm
from .report import calc_stats, report_log, report_perf

class Universe(object):
    def __init__(self, name, prices=None):
        self.name = name
        if prices is not None:
            _prices = prices.copy() 
            _prices.index = pd.to_datetime(_prices.index)
            pr = _prices.stack().rename('price')
            dr = _prices.pct_change().stack().rename('return')
            pricing = pd.concat([pr, dr], axis=1).sort_index()
            self._pricing = pricing
            self._calendar = self._build_calendar(pricing)
        else:
            self._pricing = None
            self._calendar = None
        self.blind_after = None

    def __repr__(self):
        return f'<UNIVERSE> {self.name}'

    @property
    def pricing(self):
        if self._pricing is None:
            return None
        else:
            return self._pricing.loc[:self.blind_after]

    @property
    def calendar(self):
        if self._calendar is None:
            return None
        else:
            return self._calendar.loc[:self.blind_after]

    def _build_calendar(self, pricing):
        bd = pricing.index.get_level_values(0).unique().sort_values()
        start, end = bd[0], bd[-1]
        bsd = bd.to_frame(name='TD')
        bsd['ND'] = bsd.shift(-1) # next business day
        bsd.loc[end, 'ND'] = bsd.loc[end, 'TD'] + pd.Timedelta(days=1) # fill NaN with next day
        bsd['EOD'] = True
        bsd['EOM'] = bsd.apply(lambda r: True if r['TD'].month != r['ND'].month else False, axis=1)
        bsd['EOQ'] = bsd.apply(lambda r: True if r['TD'].month != r['ND'].month and r['TD'].month in (3, 6, 9, 12) else False, axis=1)
        bsd['EOH'] = bsd.apply(lambda r: True if r['TD'].month != r['ND'].month and r['TD'].month in (6, 12) else False, axis=1)
        bsd['EOY'] = bsd.apply(lambda r: True if r['TD'].month != r['ND'].month and r['TD'].month == 12 else False, axis=1)
        cal = pd.date_range(start, end).to_frame(name='CD') # build calendar 
        cal = pd.concat([cal, bsd], axis=1)
        cal = cal[['EOD', 'EOM', 'EOQ', 'EOH', 'EOY']].astype('boolean').fillna(False)   # fill holidays with False
        return cal

    def add_pricing(self, data):
        pricing = self.pricing.append(data).sort_index()
        pr = pricing['price'].unstack().fillna(method='ffill').stack().rename('price')
        dr = pricing['return'].unstack().fillna(0).stack().rename('return')
        self._pricing = pd.concat([pr, dr], axis=1)
        self._calendar = self._build_calendar(self._pricing)

    def set_blind_after(self, date):
        self.blind_after = date


class Scheduler(object):
    def __init__(self, calendar, rule_or_list='EOM'):
        self._business_days = calendar[calendar.EOD == 1]

        if isinstance(rule_or_list, list):
            self._rule = 'LIST'
            self._rebalance_dates = pd.DatetimeIndex(rule_or_list)
        elif isinstance(rule_or_list, pd.DatetimeIndex):
            self._rule = 'LIST'
            self._rebalance_dates = rule_or_list
        elif isinstance(rule_or_list, str):  # EOM, EOQ,..
            self._rule = rule_or_list
            stdday = self._rule[:3]
            offset = 0 if len(self._rule) == 3 else int(self._rule[3:])
            self._rebalance_dates = self._business_days[self._business_days.shift(offset)[stdday] == 1].index
        else:
            raise ValueError('Input must be a keyword or a list of date')

    def __repr__(self):
        return f'<Scheduler>\nRebalance Rule: {self._rule}\nRebalance Dates: {self._rebalance_dates}'

    @property
    def rebalance_dates(self):
        return self._rebalance_dates

    def business_days(self, start, end):
        return self._business_days.loc[start:end].index

    def is_rebalance_date(self, date, *args, **kwargs):
        if self.rebalance_dates is not None:   # periodic rebalancing
            return date in self._rebalance_dates
        else:                                  # threshold rebalancing
            # not implemented
            return False


class Fund(object):
    def __init__(self):
        self.is_initiated = False
        self._nav = 100
        self._weights = pd.Series(None, dtype=float).rename('weight')

    def __repr__(self):
        return f'NAV: {self.nav}\nWeights:\n{self.weights.to_string()}'

    @property
    def nav(self):
        return self._nav

    @property
    def weights(self):
        return self._weights

    def rebalance(self, weights):
        if weights is not None:
            self.is_initiated = True
            new_portfolio = weights.rename('weight')
            old_portfolio = self._weights
            _trades = new_portfolio.subtract(old_portfolio, fill_value=0).rename('trade')
            self._weights = new_portfolio
        else:
            _trades = None
        return _trades

    def update(self, date, pricing, logger):
        _pricing = pricing.xs(date)
        _portfolio_return = np.float64(0)
        _asset_returns = []
        for asset in self._weights.index:
            # Price does not exist in the universe: cash out
            try:
                _asset_return = _pricing.loc[asset, 'return']
            except KeyError:
                logger.write('No price data', date, f'{asset} : cash out')
                self._weights.drop(asset, inplace=True)
                continue
            # To check if there are abnormal data.
            if math.isnan(_asset_return):
                logger.write('Daily return is NaN', date, f'{asset} : changed to zero')
                _asset_return = 0
            if abs(_asset_return) > 0.30:
                logger.write('Large price change', date, f'{asset} : {_asset_return:.2%}')
            _asset_returns.append([asset, _asset_return])
            _portfolio_return += self._weights.loc[asset] * _asset_return
        _asset_returns = pd.DataFrame(_asset_returns, columns=['asset', 'return']).set_index('asset')
        self._nav *= (1 + _portfolio_return)
        self._weights = self._weights.mul((1 + _asset_returns['return']), axis=0) / (1 + _portfolio_return)
        self._weights = self._weights.rename('weight')
        return _portfolio_return


class Rule(metaclass=ABCMeta):
    @abstractmethod
    def calculate(self, date, universe, fund):
        pass


class BacktestLogger(object):
    def __init__(self):
        self._log = []

    def initialize(self):
        self._log = []

    def write(self, event, date, message):
        self._log.append([event, date, message])

    def finalize(self):
        self._log = pd.DataFrame(self._log, columns=['event', 'date', 'message'])


class Portfolio(object):
    def __init__(self, name, universe, schedule, rule, cost=0):
        self.name = name
        self.universe = universe
        self.scheduler = Scheduler(universe._calendar, schedule)
        self.rule = rule
        self.cost = cost

        self.gross_returns = pd.Series(None, dtype=float).rename('return')
        self.returns = None
        self.weights = None
        self.trades = None
        self.stats = None
        self._logger = BacktestLogger()

    def __repr__(self):
        return self.name

    @property
    def log(self):
        return self._logger._log

    def _record(self, date, returns, weights, trades):
        def _to_multiindex_with_date(dt, sr):
            df = sr.to_frame().reset_index()
            df.columns = ['asset'] + [sr.name]
            df['date'] = dt
            df.set_index(['date', 'asset'], inplace=True)
            return df[sr.name]
        self.gross_returns.loc[date] = returns
        self.weights = pd.concat([self.weights, _to_multiindex_with_date(date, weights)])
        if trades is not None:
            self.trades = pd.concat([self.trades, _to_multiindex_with_date(date, trades)])

    def _calc_net_returns(self):
        # compute and subtract transaction cost from the portfolio returns
        if self.trades is None:
            self.returns = self.gross_returns.copy()
        else:
            turnover = self.trades.abs().groupby(self.trades.index.get_level_values(0)).sum().astype(float)
            self.returns = self.gross_returns.sub(turnover * self.cost, fill_value=0)
            self.stats = None # initialize for re-adjustment

    def _evaluate(self):
        self.stats = calc_stats(self.returns, self.trades)
        return self.stats

    def update(self, item, target=None, value=None):
        if item == 'rebalance_date':
            self.scheduler._rebalance_dates = self.scheduler._rebalance_dates.map(
                lambda d: pd.Timestamp(value) if d == pd.Timestamp(target) else d
            )
        else:
            raise ValueError('Not defined')

    def report(self, start='1900-01-01', end='2099-01-01', benchmark=None, relative=False, charts='interactive'):
        '''
        Report performance metrics and charts.

        Parameters
        ----------
        start : string 'YYYY-MM-DD' or datetime
        end : string 'YYYY-MM-DD' or datetime
        benchamrk : pd.Series
            daily index value or price, not daily return
        relative : boolean
            If True, excess returns will be analyzed.
        '''
        rtns = self.returns
        g_rtns = self.gross_returns
        wgts = self.weights
        if benchmark is not None:
            bm = benchmark.pct_change()
            bm.index = pd.to_datetime(bm.index)
            bm = bm.reindex(rtns.index).fillna(0)
            t0 = rtns.index[0]
            if rtns.loc[t0] == 0:
                bm.loc[t0] = 0 # To be aligned with portfolio return
            if relative:
                rtns = rtns - bm
        else:
            bm = None
        trds = self.trades

        report_perf(rtns, g_rtns, trds, wgts, bm, charts)

    def backtest(self, start='1900-01-01', end='2099-12-31', initial_weights=None, verbose=True):
        # initialize for re-run
        self.returns = pd.Series(None, dtype=float).rename('return')
        self.weights = None
        self.trades = None
        self.stats = None
        self.universe.set_blind_after(None) 
        self._logger.initialize()

        fund = Fund()
        fund.rebalance(initial_weights)

        bar_format='{percentage:3.0f}% {bar} ({desc}) {n_fmt}/{total_fmt} | \
                    Elapsed {elapsed} | Remaining {remaining} | {rate_inv_fmt}'
        business_days_iterator = tqdm(self.scheduler.business_days(start, end),
                                      bar_format=bar_format,
                                      desc='DATE', disable=(not verbose))
        for td in business_days_iterator:
            business_days_iterator.desc = td.strftime('%Y-%m-%d')

            fund_return = fund.update(td, self.universe._pricing, self._logger)

            # Rebalance
            if self.scheduler.is_rebalance_date(td):
                self.universe.set_blind_after(td) # prevent look-ahead bias 방지
                weights = self.rule.calculate(td, self.universe, fund)
                trades = fund.rebalance(weights)
                if trades is not None:
                    self._logger.write('Rebalancing', td, f'{len(trades)} trades')
            else:
                trades = None

            if fund.is_initiated:
                self._record(td, fund_return, fund.weights, trades)

        business_days_iterator.close()

        self._calc_net_returns()
        self._evaluate()
        self._logger.finalize()

        if verbose:
            report_log(self._logger._log)




import pandas as pd
from .components import Universe, Rule, Portfolio
from .rules import EqualWeight, RiskParity, ConstantWeight


def run_backtest(prices, schedule, rule, cost=0, start='1900-01-01', end='2099-12-31', verbose=True):
    '''
    Run backtest.

    Parameters
    ----------
    prices : pd.DataFrame
        Daily price data of the assets.
    schedule : string('EOM', 'EOQ',...) or list
    rule : string('EqualWeight', 'RiskParity') or dictionary or class instance
    start : string 'YYYY-MM-DD' or datetime
    end : string 'YYYY-MM-DD' or datetime
    verbose : boolean

    Returns
    -------
    portfolio : Portfolio object
        Contains the universe data, backtest policies and the backtest result.
    '''

    universe = Universe('Universe', prices)

    if rule == 'EqualWeight':
        rule = EqualWeight(prices.columns)
    elif rule == 'RiskParity':
        rule = RiskParity(prices.columns)
    elif isinstance(rule, dict):
        rule = ConstantWeight(rule)
    elif issubclass(type(rule), Rule):
        pass
    else:
        raise ValueError('rule is not valid.')

    portfolio = Portfolio('Portfolio', universe, schedule, rule, cost)
    portfolio.backtest(start=start, end=end, verbose=verbose)
    return portfolio


from setuptools import setup, find_packages
import versioneer

def readme():
    with open('README.md', encoding='utf-8') as f:
        contents = f.read()
    return contents


setup(
    name='Snowball',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Hong Junhyoung',
    author_email='bellus77@gmail.com',
    description='Investment strategy backtesting tool',
    long_description=readme(),
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/HongJunhyoung/Snowball',
    packages=find_packages(include=["snowball", "snowball.*"]),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
        'numpy>=1.21.0',
        'pandas>=1.3.5',
        'scipy>=1.7.3',
        'PyPortfolioOpt>=1.5.1',
        'tqdm>=4.45.0',
        'plotly>=5.1.0',
        'kaleido>=0.2.1',
        'Ipython'
    ],
    classifiers=[
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pandas as pd
from plotly.offline import iplot
import plotly.graph_objs as go
import plotly.io as pio
from IPython.display import display, Markdown, HTML, Image

pio.templates.default='none' # plotly style


def calc_stats(returns, trades=None):
    '''
    Calculate stats of the portfolio.

    Parameters
    ----------
    returns : pd.Series
        daily return of portfolio

    Returns
    -------
    stats : dictionary
        MTD, YTD, 1/5/10year return, Total Return, CAGR, Vol, Sharpe Ratio, MDD, etc.
    '''
    last_day = returns.index[-1]
    nav = (1 + returns).cumprod()

    stats = dict()

    t0 = last_day.replace(day=1) - pd.Timedelta(days=1) # end of previous month
    stats['MTD'] = nav.iloc[-1] / nav[:t0].iloc[-1] - 1 if not nav[:t0].empty else np.nan

    t0 = last_day.replace(month=1, day=1) - pd.Timedelta(days=1) # end of previous year
    stats['YTD'] = nav.iloc[-1] / nav[:t0].iloc[-1] - 1 if not nav[:t0].empty else np.nan

    years = [1, 5, 10]
    for n in years:
        t0 = last_day - pd.Timedelta(days=(365*n + 1)) # n years ago
        stats[f'{n}Y'] = nav.iloc[-1] / nav[:t0].iloc[-1] - 1 if not nav[:t0].empty else np.nan

    stats['Total Return'] = nav.iloc[-1] - 1

    days_per_year = 252
    stats['CAGR'] = nav.iloc[-1] ** (days_per_year / len(returns)) - 1
    stats['Volatility'] = returns.std() * np.sqrt(252)  # default ddof = 1
    stats['Sharpe Ratio'] = returns.mean() * days_per_year / stats['Volatility']

    running_max = np.maximum.accumulate(nav)
    drawdown = -((running_max - nav) / running_max)
    stats['MDD'] = np.min(drawdown)
    stats['MDD Date'] = drawdown.idxmin().strftime('%Y-%m-%d')

    # Monthly return metrics
    if returns.iloc[0] == 0:
        # Exclude initial rebalancing date
        _returns = returns.iloc[1:]
    else:
        _returns = returns
    mr = _returns.groupby([_returns.index.year, _returns.index.month]).apply(lambda x: (1+x).cumprod().iloc[-1] - 1)
    stats['Best Month'] = mr.max()
    stats['Worst Month'] = mr.min()
    stats['Positive Months'] = f'{len(mr[mr>0])} out of {len(mr)}'

    # Annual turnover
    if trades is not None:
        turnover = trades.abs().groupby(trades.index.get_level_values(0)).sum()
        turnover = turnover.iloc[1:]  # Exclude initial trades
        average_monthly_turnover = turnover.sum() / len(mr) 
        stats['Annual Turnover'] = average_monthly_turnover * 12
    else:
        stats['Annual Turnover'] = np.nan

    return stats


def report_log(log):
    '''
    Display the backtest log.

    Parameters
    ----------
    log : components.BacktestLog
        The object contains bactest log
    '''
    display(Markdown('__[ Backtest Log ]__'))
    report = []
    for event in log['event'].unique():
        _df = log[log['event']==event] 
        cnt = _df.shape[0]
        for i, row in enumerate(_df[['date', 'message']].values):
            if cnt > 5 and 2 < i < (cnt - 2):
                continue
            elif cnt > 5 and i == 2:
                report.append(['\u00B7' * 3, ' ', '  ' + '\u00B7' * 5, '\u00B7' * 5]) 
            elif i == 0:
                report.append([str(i+1), event, row[0].strftime('%Y-%m-%d'), row[1]])
            else:
                report.append([str(i+1), ' ', row[0].strftime('%Y-%m-%d'), row[1]])
    report = pd.DataFrame(report, columns=['No', 'Event', 'Date', 'Message'])
    display(HTML(report.to_html(index=False)))
    print('\n')


def report_perf(returns, gross_returns=None, trades=None, weights=None, benchmark=None, charts='interactive'):
    '''
    Report the portfolio performance.

    Parameters
    ----------
    returns : pd.Series
    gross_returns : pd.Series
    trades : pd.Series
    weights : pd.Series
    benchmark : pd.Series
    '''
    start = returns.index[0]
    end = returns.index[-1]
    months = (end.year - start.year) * 12 + end.month - start.month
    yr = months // 12
    mn = months % 12
    period = f'{yr} years' if yr >= 1 else ''
    period += f' {mn} months' if mn > 0 else ''
    stats = calc_stats(returns, trades)
    stats = pd.DataFrame.from_dict(stats, orient='index', columns=['Portfolio'])
    if benchmark is not None:
        bm_stats = calc_stats(benchmark)
        bm_stats = pd.DataFrame.from_dict(bm_stats, orient='index', columns=['Benchmark'])
        stats = pd.concat([stats, bm_stats], axis=1)

    if charts:
        display(Markdown('__[ Porfolio Performance ]__'))

    display(Markdown(f"- _{start.strftime('%Y-%m-%d')} ~ {end.strftime('%Y-%m-%d')} ( {period} )_"))
    display((stats.T
             .style.format('{:+.2%}', na_rep='N/A') # .set_properties(**{'width': '50px'}) 
                   .format('{:.2f}', subset=['Sharpe Ratio'])
                   .format('{:s}', subset=['MDD Date'])
                   .format('{:s}', subset=['Positive Months'])
                   .format('{:.2%}', subset=['Volatility', 'Annual Turnover'], na_rep='N/A')
           ))

    if charts:
        # draw charts
        fig1 = make_history_chart(returns, gross_returns, trades, weights, benchmark)
        fig2 = make_periodic_chart(returns)
        if charts == 'interactive':
            iplot(fig1)
            iplot(fig2)
        elif charts == 'static':
            display(Image(fig1.to_image(format='png', engine='kaleido', width=850, height=800)))
            display(Image(fig2.to_image(format='png', engine='kaleido', width=850, height=400)))
        else:
            raise ValueError('Charts should be interactive or static.')


def make_history_chart(returns, gross_returns, trades, weights, benchmark):
    nav = (1 + returns).cumprod() * 100
    if gross_returns is not None:
        gross_nav = (1 + gross_returns).cumprod() * 100
    running_max = np.maximum.accumulate(nav)
    drawdown = -100 * ((running_max - nav) / running_max)
 
    data = []
    data.append(go.Scatter(x=returns.index, 
                           y=nav.round(2), 
                           name='Portfolio', 
                           opacity=0.8, 
                           yaxis='y'))

    if benchmark is not None:
        bm = ((1 + benchmark).cumprod() * 100)
        data.append(go.Scatter(x=benchmark.index, 
                            y=bm.round(2), 
                            name='Benchmark', 
                            opacity=0.8, 
                            line=dict(color='grey'),
                            yaxis='y'))
    data.append(go.Bar(x=drawdown.index, 
                       y=drawdown.round(2), 
                       name='drawdown', 
                       opacity=0.8, 
                       yaxis='y2', 
                       showlegend=False))
    data.append(go.Bar(x=returns.index, 
                       y=(returns * 100).round(2), 
                       name='daily return', 
                       opacity=0.8, 
                       yaxis='y3', 
                       showlegend=False, 
                       marker_color='rgb(169,169,169)'))
    if weights is not None:
        weights = weights.loc[trades.index.get_level_values(0).tolist(), :] # Show only rebalanced weights
        df = weights.unstack()
        for col in df.columns:
            data.append(go.Bar(x=df.index, 
                               y=(df[col] * 100).round(2), 
                               name=str(col), 
                               opacity=0.8, 
                               yaxis='y4', 
                               showlegend=False))
    if trades is not None:
        # Do not draw initial trades which are not turnover.
        first_rebal_dt = trades.index.get_level_values(0).unique()[1]
        trades = trades.loc[first_rebal_dt:]
        buy = trades[trades>0].groupby('date').sum().astype(float)
        sell = trades[trades<0].groupby('date').sum().astype(float)
        data.append(go.Bar(x=buy.index, 
                           y=(buy * 100).round(2), 
                           name='Buy', 
                           opacity=0.8, 
                           yaxis='y5', 
                           showlegend=False, 
                           marker_color='rgb(244,109,67)'))
        data.append(go.Bar(x=sell.index, 
                           y=(sell * 100).round(2), 
                           name='Sell', 
                           opacity=0.8, 
                           yaxis='y5', 
                           showlegend=False, 
                           marker_color='rgb(69,117,180)'))

    if gross_returns is not None:
        data.append(go.Scatter(x=gross_returns.index,
                               y=gross_nav.round(2),
                               name='Before cost',
                               opacity=0.3, line_color='gray',
                               yaxis='y'))

    yaxis_lin = dict(title='NAV (Linear scale)', autorange=True, domain=[0.50, 1])
    yaxis_log = dict(title='NAV (Log scale)', autorange=True, domain=[0.50, 1], type='log')

    layout = go.Layout(
        height=1000,
        margin=go.layout.Margin(b=30, r=40),
        xaxis=dict(
            title='',
            rangeselector=dict(
                buttons=list([
                    dict(count=1, label='1M', step='month', stepmode='backward'),
                    dict(count=3, label='3M', step='month', stepmode='backward'),
                    dict(count=6, label='6M', step='month', stepmode='backward'),
                    dict(count=1, label='1Y', step='year', stepmode='backward'),
                    dict(count=3, label='3Y', step='year', stepmode='backward'),
                    dict(count=5, label='5Y', step='year', stepmode='backward'),
                    dict(count=10, label='10Y', step='year', stepmode='backward'),
                    dict(count=15, label='15Y', step='year', stepmode='backward'),
                    dict(count=20, label='20Y', step='year', stepmode='backward'),
                    dict(count=1, label='MTD', step='month', stepmode='todate'),
                    dict(count=1, label='YTD', step='year', stepmode='todate'),
                    ])),
            type='date',
            autorange=True,
            showspikes=True,
        ),
        yaxis=yaxis_log,
        yaxis2=dict(
            title='drawdown(%)',
            autorange=True,
            domain=[0.36, 0.46],
            anchor='x'
        ),
        yaxis3=dict(
            title='return(%)',
            autorange=True,
            domain=[0.24, 0.34],
            anchor='x'
        ),
        yaxis4=dict(
            title='weights(%)',
            autorange=True,
            domain=[0.12, 0.22],
            anchor='x'
        ),
        yaxis5=dict(
            title='trades(%)',
            autorange=True,
            domain=[0.00, 0.10],
            anchor='x'
        ),
        barmode='relative',
        updatemenus=list([
            dict(type='buttons',
                 active=0,
                 buttons=list([
                     dict(label='Log',
                          method='update',
                          args=[{}, {'yaxis': yaxis_log}]),
                     dict(label='Linear',
                          method='update',
                          args=[{}, {'yaxis': yaxis_lin}]),
                 ]),
            )
        ]),

        legend=dict(orientation='h', xanchor='right', x=1, yanchor='bottom', y=1.01)
    )

    figure = go.Figure(data=data, layout=layout)
    return figure

def make_periodic_chart(returns):
    # Annual returns 
    yr = returns.groupby(returns.index.year).apply(lambda x: (1+x).cumprod().iloc[-1] - 1)
    trace_annual = go.Bar(x=yr * 100,
                          y=yr.index,
                          name='Annual Return',
                          marker=dict(color='#7EC0EE'),
                          orientation='h',
                          hoverinfo='text',
                          hovertext=[f'{y:n}   {v:+.2%}' for y, v in yr.reset_index().values],
                          xaxis='x'
                          )
    trace_mean = go.Scatter(x=[yr.mean() * 100] * len(yr.index),
                            y=yr.index,
                            name='Average',
                            line=dict(dash='dot'),
                            hoverinfo='text',
                            hovertext=[f'{yr.mean():+.2%}'] * len(yr.index),
                            xaxis='x'
                            )

    # Monthly returns
    mr = returns.groupby([returns.index.year, returns.index.month]).apply(lambda x: (1+x).cumprod().iloc[-1] - 1)
    hm = mr.unstack() # unstack months to columns
    x = hm.columns # month
    y = hm.index   # year
    z = (hm * 100).values.tolist()
    # hoverinfo
    text = []
    for year_l_index, year_l in enumerate(hm.values.tolist()):
        text_temp = []
        for month_index, value in enumerate(year_l):
            yyyymm = str(hm.index.tolist()[year_l_index]) + '-' + str(hm.columns.tolist()[month_index]).rjust(2, '0')
            text_temp.append(f'{yyyymm}   {value:+.2%}')
        text.append(text_temp)
    colorscale = [
        [0.0, 'rgb(49,54,149)'],
        [0.1111111111111111, 'rgb(69,117,180)'],
        [0.2222222222222222, 'rgb(116,173,209)'],
        [0.3333333333333333, 'rgb(171,217,233)'],
        [0.4444444444444444, 'rgb(224,243,248)'],
        [0.5555555555555555, 'rgb(254,224,144)'],
        [0.6666666666666666, 'rgb(253,174,97)'],
        [0.7777777777777777, 'rgb(244,109,67)'],
        [0.8888888888888888, 'rgb(215,48,39)'],
        [1.0, 'rgb(165,0,38)']]
    sigma = np.std(mr.dropna().values * 100)
    trace_monthly = go.Heatmap(x=x,
                               y=y,
                               z=z,
                               zauto=False,
                               zmin=-3*sigma,
                               zmax=3*sigma,
                               hoverinfo='text',
                               hovertext=text,
                               colorscale=colorscale,
                               showscale=True,
                               xaxis='x2',
    )

    layout = go.Layout(height=400,
                       margin=go.layout.Margin(t=25, l=120),
                       xaxis=dict(title='Annual return', domain=[0.00, 0.4]),
                       xaxis2=dict(title='Monthly return', domain=[0.45, 1.0]),
                       yaxis=dict(title='Year', autorange='reversed', tickformat='d'),
                       hovermode='closest',
                       autosize=True,  # 사이즈 자동
                       showlegend=False,
    )

    data = [trace_annual, trace_mean, trace_monthly]
    figure = go.Figure(data=data, layout=layout)
    return figure
